# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - MoinMessage library

    @copyright: 2012, 2013, 2014 by Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from email import message_from_string
from email.encoders import encode_noop
from email.generator import Generator
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.parser import Parser
from email.utils import formatdate
from itertools import islice
from subprocess import Popen, PIPE
from tempfile import mkstemp
from urlparse import urlsplit
from DateSupport import getDateTimeFromRFC2822
import httplib
import os

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

# Message inspection functions.

def is_collection(message):
    return message.get("Update-Type") == "collection"

def to_replace(message):
    return message.get("Update-Action") == "replace"

def to_store(message):
    return message.get("Update-Action") == "store"

def get_update_action(message):
    return message.get("Update-Action", "update")

# Core abstractions.

class Message:

    "An update message."

    def __init__(self, text=None):
        self.date = None
        self.updates = []
        if text:
            self.parse_text(text)

    def init_date(self, message):

        "Obtain the date of the given 'message'."

        if message.has_key("Date"):
            self.date = getDateTimeFromRFC2822(message["Date"])
        else:
            self.date = None

    def parse_text(self, text):

        "Parse the given 'text' as a message."

        self.handle_message(message_from_string(text))

    def handle_message(self, message):

        "Handle the given 'message', recording the separate updates."

        self.init_date(message)

        # The message either consists of a collection of updates.

        if message.is_multipart() and is_collection(message):
            for part in message.get_payload():
                self.updates.append(part)

        # Or the message is a single update.

        else:
            self.updates.append(message)

    def add_updates(self, parts):

        """
        Add the given 'parts' to a message.
        """

        for part in updates:
            self.add_update(part)

    def add_update(self, part):

        """
        Add an update 'part' to a message.
        """

        self.updates.append(part)

    def get_update(self, alternatives):

        """
        Return a suitable multipart object containing the supplied
        'alternatives'.
        """

        part = MIMEMultipart("alternative")
        for alternative in alternatives:
            part.attach(alternative)
        return part

    def get_payload(self, subtype="mixed", timestamped=True):

        """
        Get the multipart payload for the message. If the 'timestamped'
        parameter is omitted or set to a true value, the payload will be given a
        date header set to the current date and time that can be used to assess
        the validity of a message and to determine whether it has already been
        received by a recipient.
        """

        if len(self.updates) == 1:
            message = self.updates[0]
        else:
            message = MIMEMultipart(subtype)
            message.add_header("Update-Type", "collection")
            for update in self.updates:
                message.attach(update)

        if timestamped:
            timestamp(message)
            self.init_date(message)

        return message

class MoinMessageError(Exception):
    pass

class MoinMessageDecodingError(Exception):
    pass

class MoinMessageMissingPart(MoinMessageDecodingError):
    pass

class MoinMessageBadContent(MoinMessageDecodingError):
    pass

class GPG:

    "A wrapper around the gpg command using a particular configuration."

    def __init__(self, homedir=None):
        self.conf_args = []

        if homedir:
            self.conf_args += ["--homedir", homedir]

        self.errors = None

    def run(self, args, text=None):

        """
        Invoke gpg with the given 'args', supplying the given 'text' to the
        command directly or, if 'text' is omitted, using a file provided as part
        of the 'args' if appropriate.

        Failure to complete the operation will result in a MoinMessageError
        being raised.
        """

        cmd = Popen(["gpg"] + self.conf_args + list(args), stdin=PIPE, stdout=PIPE, stderr=PIPE)

        # Attempt to write input to the command and to read output from the
        # command.

        text, self.errors = cmd.communicate(text)

        # Test for a zero result.

        if not cmd.returncode:
            return text
        else:
            raise MoinMessageError, self.errors

    def verifyMessageText(self, signature, content):

        "Using the given 'signature', verify the given message 'content'."

        # Write the detached signature and content to files.

        signature_fd, signature_filename = mkstemp()
        content_fd, content_filename = mkstemp()

        try:
            signature_fp = os.fdopen(signature_fd, "w")
            content_fp = os.fdopen(content_fd, "w")
            try:
                signature_fp.write(signature)
                content_fp.write(content)
            finally:
                signature_fp.close()
                content_fp.close()

            # Verify the message text.

            text = self.run(["--status-fd", "1", "--verify", signature_filename, content_filename])

            # Return the details of the signing key.

            identity = None
            fingerprint = None

            for line in text.split("\n"):
                try:
                    prefix, msgtype, digest, details = line.strip().split(" ", 3)
                except ValueError:
                    continue

                # Return the fingerprint and identity details.

                if msgtype == "GOODSIG":
                    identity = details
                elif msgtype == "VALIDSIG":
                    fingerprint = digest

                if identity and fingerprint:
                    return fingerprint, identity

            return None

        finally:
            os.remove(signature_filename)
            os.remove(content_filename)

    def verifyMessage(self, message):

        """
        Verify the given RFC 3156 'message', returning a tuple of the form
        (fingerprint, identity, content).
        """

        content, signature = getContentAndSignature(message)

        # Verify the message format.

        if signature.get_content_type() != "application/pgp-signature":
            raise MoinMessageBadContent

        # Verify the message.

        fingerprint, identity = self.verifyMessageText(signature.get_payload(decode=True), as_string(content))
        return fingerprint, identity, content

    def signMessage(self, message, keyid):

        """
        Return a signed version of 'message' using the given 'keyid'.
        """

        # Sign the container's representation.

        signature = self.run(["--armor", "-u", keyid, "--detach-sig"], as_string(message))

        # Make the container for the message.

        signed_message = MIMEMultipart("signed", protocol="application/pgp-signature")
        signed_message.attach(message)

        signature_part = MIMEBase("application", "pgp-signature")
        signature_part.set_payload(signature)
        signed_message.attach(signature_part)

        return signed_message

    def decryptMessageText(self, message):

        "Return a decrypted version of 'message'."

        return self.run(["--decrypt"], message)

    def decryptMessage(self, message):

        """
        Decrypt the given RFC 3156 'message', returning the message text.
        """

        try:
            declaration, content = message.get_payload()
        except ValueError:
            raise MoinMessageMissingPart

        # Verify the message format.

        if content.get_content_type() != "application/octet-stream":
            raise MoinMessageBadContent

        # Return the decrypted message text.

        return self.decryptMessageText(content.get_payload(decode=True))

    def encryptMessage(self, message, keyid):

        """
        Return an encrypted version of 'message' using the given 'keyid'.
        """

        text = as_string(message)
        encrypted = self.run(["--armor", "-r", keyid, "--encrypt", "--trust-model", "always"], text)

        # Make the container for the message.

        encrypted_message = MIMEMultipart("encrypted", protocol="application/pgp-encrypted")

        # For encrypted content, add the declaration and content.

        declaration = MIMEBase("application", "pgp-encrypted")
        declaration.set_payload("Version: 1")
        encrypted_message.attach(declaration)

        content = MIMEApplication(encrypted, "octet-stream", encode_noop)
        encrypted_message.attach(content)

        return encrypted_message

    def importKeys(self, text):

        """
        Import the keys provided by the given 'text'.
        """

        self.run(["--import"], text)

    def exportKey(self, keyid):

        """
        Return the "armoured" public key text for 'keyid' as a message part with
        a suitable media type.
        See: https://tools.ietf.org/html/rfc3156#section-7
        """

        text = self.run(["--armor", "--export", keyid])
        return MIMEApplication(text, "pgp-keys", encode_noop)

    def listKeys(self, keyid=None):

        """
        Return a list of key details for keys on the keychain, selecting only
        one specific key if 'keyid' is specified.
        """

        text = self.run(["--list-keys", "--with-colons", "--with-fingerprint"] +
            (keyid and ["0x%s" % keyid] or []))
        return self._getKeysFromResult(text)

    def listSignatures(self, keyid=None):

        """
        Return a list of key and signature details for keys on the keychain,
        selecting only one specific key if 'keyid' is specified.
        """

        text = self.run(["--list-sigs", "--with-colons", "--with-fingerprint"] +
            (keyid and ["0x%s" % keyid] or []))
        return self._getKeysFromResult(text)

    def getKeysFromMessagePart(self, part):

        """
        Process an application/pgp-keys message 'part', returning a list of
        key details.
        """

        return self.getKeysFromString(part.get_payload(decode=True))

    def getKeysFromString(self, s):

        """
        Return a list of key details extracted from the given key block string
        's'. Signature information is also included through the use of the gpg
        verbose option.
        """

        text = self.run(["--with-colons", "--with-fingerprint", "-v"], s)
        return self._getKeysFromResult(text)

    def _getKeysFromResult(self, text):

        """
        Return a list of key details extracted from the given command result
        'text'.
        """

        keys = []
        for line in text.split("\n"):
            try:
                recordtype, trust, keylength, algorithm, keyid, cdate, expdate, serial, ownertrust, _rest = line.split(":", 9)
            except ValueError:
                continue

            if recordtype == "pub":
                userid, _rest = _rest.split(":", 1)
                keys.append({
                    "type" : recordtype, "trust" : trust, "keylength" : keylength,
                    "algorithm" : algorithm, "keyid" : keyid, "cdate" : cdate,
                    "expdate" : expdate, "userid" : userid, "ownertrust" : ownertrust,
                    "fingerprint" : None, "subkeys" : [], "signatures" : []
                    })
            elif recordtype == "sub" and keys:
                keys[-1]["subkeys"].append({
                    "trust" : trust, "keylength" : keylength, "algorithm" : algorithm,
                    "keyid" : keyid, "cdate" : cdate, "expdate" : expdate,
                    "ownertrust" : ownertrust
                    })
            elif recordtype == "fpr" and keys:
                fingerprint, _rest = _rest.split(":", 1)
                keys[-1]["fingerprint"] = fingerprint
            elif recordtype == "sig" and keys:
                userid, _rest = _rest.split(":", 1)
                keys[-1]["signatures"].append({
                    "keyid" : keyid, "cdate" : cdate, "expdate" : expdate,
                    "userid" : userid
                    })

        return keys

# Message serialisation functions, working around email module problems.

def as_string(message):

    """
    Return the string representation of 'message', attempting to preserve the
    precise original formatting.
    """

    out = StringIO()
    generator = Generator(out, False, 0) # disable reformatting measures
    generator.flatten(message)
    return out.getvalue()

# Message decoding functions.

# Detect PGP/GPG-encoded payloads.
# See: http://tools.ietf.org/html/rfc3156

def is_signed(message):
    mimetype = message.get_content_type()
    encoding = message.get_content_charset()

    return mimetype == "multipart/signed" and \
        message.get_param("protocol") == "application/pgp-signature"

def is_encrypted(message):
    mimetype = message.get_content_type()
    encoding = message.get_content_charset()

    return mimetype == "multipart/encrypted" and \
        message.get_param("protocol") == "application/pgp-encrypted"

def getContentAndSignature(message):

    """
    Return the content and signature parts of the given RFC 3156 'message'.

    NOTE: RFC 3156 states that signed messages should employ a detached
    NOTE: signature but then shows "BEGIN PGP MESSAGE" for signatures
    NOTE: instead of "BEGIN PGP SIGNATURE".
    NOTE: The "micalg" parameter is currently not supported.
    """

    try:
        content, signature = message.get_payload()
        return content, signature
    except ValueError:
        raise MoinMessageMissingPart

# Communications functions.

def timestamp(message):

    """
    Timestamp the given 'message' so that its validity can be assessed by the
    recipient.
    """

    datestr = formatdate()

    if not message.has_key("Date"):
        message.add_header("Date", datestr)
    else:
        message.replace_header("Date", datestr)

def _getConnection(scheme):

    "Return the connection class for the given 'scheme'."

    if scheme == "http":
        return httplib.HTTPConnection
    elif scheme == "https":
        return httplib.HTTPSConnection
    else:
        raise MoinMessageError, "Communications protocol not supported: %s" % scheme

def sendMessageOpener(message, url, method="PUT"):

    """
    Send 'message' to the given 'url' using the given 'method' (using PUT as the
    default if omitted).
    """

    scheme, host, port, path = parseURL(url)
    text = as_string(message)

    req = _getConnection(scheme)(host, port)
    req.request(method, path, text)
    resp = req.getresponse()

    if resp.status >= 400:
        raise MoinMessageError, "Message sending failed (%s): %s" % (resp.status, resp.read())

    return resp

def sendMessage(message, url, method="PUT"):

    """
    Send 'message' to the given 'url' using the given 'method' (using PUT as the
    default if omitted).
    """

    resp = sendMessageOpener(message, url, method)
    return resp.read()

def parseURL(url):

    "Return the scheme, host, port and path for the given 'url'."

    scheme, host_port, path, query, fragment = urlsplit(url)
    host_port = host_port.split(":")

    if query:
        path += "?" + query

    if len(host_port) > 1:
        host = host_port[0]
        port = int(host_port[1])
    else:
        host = host_port[0]
        port = 80

    return scheme, host, port, path

# Message handling.

class MessageInterface:

    "A command-based interface to a message store, inspired by RFC 1939 (POP3)."

    def __init__(self, store):
        self.store = store

    def execute(self, commands):

        """
        Access messages according to the given 'commands' script, acting on the
        store provided during initialisation and returning a message object
        containing the results.
        """

        # Build a container for the responses.

        message = Message()

        # Process each command.

        for command in commands.split("\n"):
            command = command.strip()

            # Get the command and arguments.

            command_parts = command.split(None, 1)
            cmd = command_parts[0]

            try:
                if cmd in self.commands:
                    getattr(self, cmd)(command_parts, message)
                else:
                    self.add_result(cmd, command, "ERR", message)
            except Exception, exc:
                self.add_result(cmd, "\n".join([command, str(exc)]), "ERR", message)

        return message

    def get_count(self, command_parts):

        # Select all messages by default.

        count = None

        if len(command_parts) > 1:
            count = int(command_parts[1])

        return count

    def add_result(self, cmd, result, status, message):
        part = MIMEText(result, "x-moinmessage-fetch-result")
        part["Request-Type"] = cmd
        part["Request-Status"] = status
        message.add_update(part)

    def add_messages(self, resources, message):
        container = Message()

        for message_text in resources:
            message_item = Parser().parsestr(message_text)
            container.add_update(message_item)

        # Convert the container to a proper multipart section.

        message.add_update(container.get_payload())

    def STAT(self, command_parts, message):

        # A request to count the messages is returned in a part.

        self.add_result("STAT", str(len(self.store)), "OK", message)

    def RETR(self, command_parts, message):

        # A request for specific messages returns each message
        # in its own part within a collection part.

        count = self.get_count(command_parts)

        self.add_messages(islice(iter(self.store), count), message)

    def DELE(self, command_parts, message):

        # A request to delete messages is performed immediately.

        count = self.get_count(command_parts)

        keys = self.store.keys()[:count]

        for key in keys:
            del self.store[key]

        self.add_result("DELE", str(len(keys)), "OK", message)

    # Command manifest, may be extended by subclasses.

    commands = "STAT", "RETR", "DELE"

# vim: tabstop=4 expandtab shiftwidth=4
